#ifdef SUNCOMPILATION
	#include <ncurses/ncurses.h>
#else
	#include <ncurses.h>
#endif
#include <string.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h> 
#include <pthread.h>
#include <signal.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>

#ifndef MSG_NOSIGNAL
  #define MSG_NOSIGNAL 0
#endif

#define BUFLEN 256
#define INBUFLEN 256
#define PORT 6666
#define CLOSMSG "/clos"
#define BLOCK_SIZE 1024
#define SLEEP_ON_EAGAIN 1
#define upprint(...) mvwprintw(mainwin, LINES-5, 1, __VA_ARGS__)

WINDOW *mainwin;
int thesock;

void upscroll() {
	scroll(mainwin);
}

void uprefresh() {
	box(mainwin, 0, 0);
	wrefresh(mainwin);
}

int send_file(int thesock, char *buf) {
	char *ptr, *token;
	fd_set fd_writeready;
	struct timeval waittime;
	token = strtok_r(buf, ",", &ptr);
	token = strtok_r(NULL, ",", &ptr);
	scroll(mainwin);
	mvwprintw(mainwin, LINES-5, 1, "Filename: %s", token);
	box(mainwin, 0, 0);
	wrefresh(mainwin);

	FILE* fd = fopen(token, "rb");
	if(fd){
		//unsigned long long size;
		size_t size;
		//char *data;
		char data[BLOCK_SIZE];
		if(fseek(fd, 0, SEEK_END) == -1) {
			fclose(fd);
			return -1;
		}
		size = ftell(fd);

		upscroll();
		upprint("Size: %llu", (unsigned long long)size);
		uprefresh();

		//data = malloc(size);

		sprintf(buf, "/file,%s, %llu,", token, (unsigned long long)size);
		send(thesock, buf, strlen(buf), MSG_NOSIGNAL);

		upscroll();

		if(fseek(fd, 0, SEEK_SET) == -1) {
			fclose(fd);
			return -1;
		}

		//unsigned long long done = 0, sentofblock = BLOCK_SIZE, blocksize = BLOCK_SIZE;
		size_t done = 0, sentofblock = BLOCK_SIZE, blocksize = BLOCK_SIZE;
		int retval, sel_retval, readblocks, eagain = 0;
		//char* block = data;
		while(done < size) {
			if(!eagain && (sentofblock == BLOCK_SIZE)) {
				blocksize = (size - done < BLOCK_SIZE) ? size - done : BLOCK_SIZE;
				readblocks = fread(data, blocksize, 1, fd);
				if(readblocks != 1) {
					if(ferror(fd)) {
						upscroll();
						upprint("File Error!");
						upscroll();
						uprefresh();
						fclose(fd);
						return -1;
					}
					if(feof(fd)) {
						upscroll();
						upprint("EOF.");
						upscroll();
						uprefresh();
					}
				}
				sentofblock = 0;
			}

			FD_ZERO(&fd_writeready);
			FD_SET(thesock, &fd_writeready);
			waittime.tv_sec = 10;
			waittime.tv_usec = 0;
			sel_retval = select(thesock + 1, NULL, &fd_writeready,NULL, &waittime);
			if(sel_retval <= 0) {
				upprint("TIMEOUT\n");
				uprefresh();
				fclose(fd);
				return -1;
			}			
			retval = send(thesock, data, readblocks*blocksize - sentofblock, /*MSG_DONTWAIT |*/ MSG_NOSIGNAL);
			/*if(retval <= 0) {
				if(!eagain && (retval == 0 || errno == EAGAIN)) {
					eagain = 1;
					upscroll();
					upprint("Sleeping.");
					upscroll();
					uprefresh();
					sleep(SLEEP_ON_EAGAIN);
				} else {*/
			if(retval < 0) {
					upprint("ERROR: %s", strerror(errno));
					uprefresh();
					fclose(fd);
					return -1;
				
			} else {
				upprint("%llu%% done, %llu, %llu", (unsigned long long)(100*done/size), (unsigned long long)(size - done), (unsigned long long)sentofblock);
				uprefresh();
				done += retval;
				sentofblock += retval;
				eagain = 0;
				//block += retval;
			}
		}

		fclose(fd);
		//free(data);
	}
	
	upscroll();
	upprint("DONE.");
	uprefresh();

	return 0;
}

void* datathread(void* param) {
	char buf[INBUFLEN], *bufptr;
	int recvcnt;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	while(1) {
		box(mainwin, 0, 0);
		wrefresh(mainwin);

		memset(buf, 0, sizeof(buf));
		recvcnt = recv(thesock, buf, INBUFLEN, 0);
		if(recvcnt > 0) {
			for(bufptr = buf; bufptr < buf + recvcnt; ++bufptr) if(*bufptr == '\n' || *bufptr == '\r') *bufptr = ' ';
			scroll(mainwin);
			mvwprintw(mainwin, LINES-5, 1, "%s", buf);
		} else {
			scroll(mainwin);
			mvwprintw(mainwin, LINES-5, 1, "No Connection");
			box(mainwin, 0, 0);
			wrefresh(mainwin);
			break;
		}
	}
	return NULL;
}

int main(int argc, char **argv) {
	WINDOW *inputwin;
	char buf[BUFLEN], *tmpbuf;
	int cmdport;

	pthread_t datathread_id;
	
  struct sockaddr_in remoteaddr;
	struct addrinfo hints, *res_addr;

	if(argc != 3) {
		printf("%s HOST PORT\n", argv[0]);
		return 1;
	}

	sscanf(argv[2], "%d", &cmdport);

	signal(SIGPIPE, SIG_IGN);

	// ======= NETWORK ======

  // The PF = Internet Family, 0 - autodefault for INET/STREAM = TCP 
  if((thesock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		printf("Cannot create a socket (%d).\n", errno);
		exit(1);
	}

  // For my own safety, clean the structure
  memset(&remoteaddr, 0, sizeof(remoteaddr));
  // Init the addr structure
  remoteaddr.sin_family = AF_INET;
  // Local to network byte order
  //remoteaddr.sin_port = htons(PORT);
  remoteaddr.sin_port = htons(cmdport);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	if(getaddrinfo(argv[1], NULL, &hints, &res_addr)) {
		printf("Cannot locate the host.\n");
		close(thesock);
		exit(1);
	}

	memcpy(&remoteaddr.sin_addr.s_addr, &((struct sockaddr_in*)res_addr->ai_addr)->sin_addr.s_addr, sizeof(remoteaddr.sin_addr.s_addr));
	freeaddrinfo(res_addr);
	//printf("IP:%u\n", ((struct sockaddr_in*)res_addr->ai_addr)->sin_addr.s_addr);
	//printf("IP:%u\n", remoteaddr.sin_addr.s_addr);
	
	// Connect
	if((connect(thesock, (struct sockaddr*)&remoteaddr, sizeof(struct sockaddr_in))) == -1) {
		printf("Cannot connect (%d).\n", errno);
		close(thesock);
		exit(1);
	}

	// ======= NCURSES ======

	memset(buf, 0, BUFLEN);

	initscr();
	refresh();

	mainwin = newwin(LINES-3, COLS, 0, 0);
	scrollok(mainwin, true);
	wsetscrreg(mainwin, 1,LINES-5);
	inputwin = newwin(3, COLS, LINES-3, 0);

	mvwprintw(mainwin, LINES-5, 1, "Connected.");

	// Data thread
	pthread_create(&datathread_id, NULL, datathread, NULL);

	while(1) {
		box(inputwin, 0, 0);
		wrefresh(inputwin);

		mvwgetnstr(inputwin, 1, 1, buf, (BUFLEN < COLS-2) ? BUFLEN : COLS -2);
		for(tmpbuf = buf; *tmpbuf != '\n' && *tmpbuf != '\0'; ++tmpbuf);
		*tmpbuf = '\0';

		if(!strcmp(buf, "/quit")) break;


		if(!strncmp(buf, "/file", 5))
			send_file(thesock, buf);
		else
			send(thesock, buf, strlen(buf), MSG_DONTWAIT | MSG_NOSIGNAL);

		wclear(inputwin);
		//scroll(mainwin);
		//mvwprintw(mainwin, LINES-5, 1, "%s", buf);
	}


	send(thesock, CLOSMSG, strlen(CLOSMSG), MSG_DONTWAIT | MSG_NOSIGNAL);

	pthread_cancel(datathread_id);

	delwin(mainwin);
	delwin(inputwin);
	endwin();

	return 0;
}
