#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <errno.h>
#include <stdlib.h>

#ifndef MSG_NOSIGNAL
	#define MSG_NOSIGNAL 0
#endif

#define LISTEN_PORT 6666
#define MAX_QUEUED_CONNS 5
#define MAX_CLIENTS 5
#define MAX_NICK_LEN 40
#define MAX_MSG_LEN 256
#define MAX_CMD_LEN 5
#define SEND_LIMIT 1000

#define DEBUG(...) //fprintf(stderr, __VA_ARGS__)
#define DEBUG2(...) fprintf(stderr, __VA_ARGS__)
#define ERROR(...) fprintf(stderr, __VA_ARGS__)
#define SEM_V(X) semop((X), &V, 1);
#define SEM_P(X) semop((X), &P, 1);

int maxsock = 0;
int sec_clients;
pthread_t datathread_id, listeningthread_id, timethread_id;
FILE *logfile;
int cmdport;

struct sembuf P = {0, -1, 0};
struct sembuf V = {0,  1, 0};

int close_client(int i);

struct client {
	int socket; // socket descriptor
	char nick[MAX_NICK_LEN+1]; // nick
	struct in_addr addr; // inet address
	uint16_t port; // port number
	int written;
} clients[MAX_CLIENTS]; // clients table

// Keep in mind the max sock number for select()
void setmaxsock(int newsock) {
	if(newsock > maxsock) maxsock = newsock;
}

// Broadcast the message
int send_to_clients(int u, char* buf, int read_retval) {
	int write_retval = 0, i, msglen;
	char tmpbuf[ MAX_NICK_LEN+MAX_MSG_LEN+15+20 ];

	memset(tmpbuf, 0, sizeof(tmpbuf));
	//sprintf(tmpbuf, "%s:%s:%u: ", clients[u].nick, inet_ntoa(clients[u].addr), ntohs(clients[u].port));
	sprintf(tmpbuf, "%s:%s: %s", clients[u].nick, inet_ntoa(clients[u].addr), buf);
	if(fprintf(logfile, "%s\n", tmpbuf) < 0)
		ERROR("Cannot write to log file!\n");
	msglen = strlen(tmpbuf);
	clients[u].written += msglen;

	if(clients[u].written > SEND_LIMIT) {
		close_client(u);
		return 0;
	}

	for(i = 0; i < MAX_CLIENTS; ++i) { // to all clients
		if(clients[i].socket) {
			write_retval |= send(clients[i].socket, tmpbuf, msglen, MSG_DONTWAIT | MSG_NOSIGNAL);
			//write_retval |= send(clients[i].socket, tmpbuf, sizeof(tmpbuf), MSG_DONTWAIT);
			//write_retval |= send(clients[i].socket, buf, read_retval, MSG_DONTWAIT);
			if(write_retval < 0 && errno != EAGAIN) { // can't wait for people with too small buffers
				DEBUG2("write retval of client %d for socket %d: %d (%d)\n", i, clients[i].socket, write_retval, errno);
				close_client(i); // clear the slot
			}
		}
	}
	return write_retval;
}

int tell_to_give_a_nick(int i) {
	const char buf[] = "Cannot send anything before you enter your nick.";
	return send(clients[i].socket, buf, strlen(buf), MSG_DONTWAIT | MSG_NOSIGNAL);
}

int set_nick(int i, char* buf) {
	int j;
	char tmpnick[MAX_NICK_LEN+1];
	char* k = tmpnick;
	const char failmsgbuf[] = "Your nick must be unique.";
	const char nickmsgbuf[] = "Nick has been set.";

	if(strlen(buf) <= MAX_CMD_LEN+1) return 1;

	strncpy(tmpnick, buf+MAX_CMD_LEN+1, MAX_NICK_LEN);
	tmpnick[MAX_NICK_LEN] = '\0';
	for(; *k; ++k)
		if(*k == '\n' || *k == '\r' || *k == ' ') *k = '\0';

	DEBUG2("tmpnick:%s\n", tmpnick);

	SEM_P(sec_clients);
		for(j = 0; j < MAX_CLIENTS; ++j) {
			if(!strcmp(tmpnick, clients[j].nick)) {
				send(clients[i].socket, failmsgbuf, strlen(failmsgbuf), MSG_DONTWAIT | MSG_NOSIGNAL);
				SEM_V(sec_clients);
				return 1;
			}
		}
	SEM_V(sec_clients);

	strcpy(clients[i].nick, tmpnick);

	send(clients[i].socket, nickmsgbuf, strlen(nickmsgbuf), MSG_DONTWAIT | MSG_NOSIGNAL);

	DEBUG("the nick:%s\n", clients[i].nick);
	return 0;
}

int close_client(int i) {
	int errcode;
	DEBUG("closing socket: %d\n", i);
	errcode = close(clients[i].socket);
	/*memset(clients[i].nick, 0, MAX_NICK_LEN+1);
	clients[i].socket = 0; // this is the last operation - slot emptiness marker*/
	// Cleaning - scanning may occur
	SEM_P(sec_clients);
		memset(&clients[i], 0, sizeof(struct client));
	SEM_V(sec_clients);
	return errcode;
}

//int recv_file(int i, char* buf, unsigned long long sentbytes) {
int recv_file(int i, char* buf, size_t sentbytes) {
	char *token, *ptr;
	//unsigned long long size;
	size_t size;
	int readret;
	char name[256];
	char data[1024];
	FILE* fd;
  fd_set fd_readready;
  struct timeval waittime;
	int sel_retval;

	// /file
	sentbytes -= strlen(token = strtok_r(buf, ",", &ptr)) + 1;
	if(!token) return -1;

	// name
	sentbytes -= strlen(token = strtok_r(NULL, ",", &ptr)) + 1;
	if(!token) return -1;

	snprintf(name, sizeof(name), "new.%s", token);

	//if((fd = fopen(name, "w"))) {
	if((fd = fopen("dump", "wb"))) {
		// size
		sentbytes -= strlen(token = strtok_r(NULL, ",", &ptr)) + 1;
		if(!token) {
			fclose(fd);
			return -1;
		}

		size = atoi(token);
		DEBUG2("SIZE: %llu SBYTES:%llu\n", (unsigned long long)size, (unsigned long long)sentbytes);

		//token = strtok_r(NULL, ",", &ptr);

		// data
		token += strlen(token)+1;

		fwrite( token, sentbytes, 1, fd);


		while(sentbytes < size) {
			/* Select */
			FD_ZERO(&fd_readready);
			FD_SET(clients[i].socket, &fd_readready);
			waittime.tv_sec = 10;
			waittime.tv_usec = 0;
			sel_retval = select(clients[i].socket + 1, &fd_readready, NULL, NULL, &waittime);
			if(sel_retval <= 0) {
				DEBUG2("TIMEOUT\n");
				fclose(fd);
				return -1;
			}
			/* ------ */
			//DEBUG2(".");
			readret = recv(clients[i].socket, data, sizeof(data), 0);
			if(readret == 0 || readret < 0) {
				DEBUG2("ERROR: %s\n", strerror(errno));
				fclose(fd);
				return -1;
			} else {
				if(fwrite( data, readret, 1, fd) != 1) {
					fclose(fd);
					return -1;
				}
				sentbytes += readret;
			}
		}
		fclose(fd);
	}

	DEBUG2("DONE.\n");
	return 0;
}

// Sanitize
int clear_buf(char *buf) {
	for(; *buf; ++buf) {
		switch(*buf) {
			case '\n':
			case '\r':
				*buf = ' ';
				break;
		}
	}
	return 0;
}

// Reading from clients
void* datathread(void* param) {
	fd_set fd_readready;
	struct timeval waittime;
	int sel_retval, read_retval, i;

	char buf[MAX_MSG_LEN + 1];
	char cmdbuf[MAX_CMD_LEN + 1];

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	signal(SIGPIPE, SIG_IGN);

	DEBUG("Data thread\n");

	while(1){
		// Clear the readready mask
		FD_ZERO(&fd_readready);
		// Mark waiting sockets
		for(i = 0; i < MAX_CLIENTS; ++i)
			if(clients[i].socket) FD_SET(clients[i].socket, &fd_readready);
		// The waittime structure needs rewriting
		//  -- select() changes it to:
		//  "the amount of time not slept"
		waittime.tv_sec = 1;
		waittime.tv_usec = 0;
		// Wait for an incoming subspace signal
		DEBUG("Before select\n");
		sel_retval = select(maxsock + 1, &fd_readready, NULL, NULL, &waittime);
		DEBUG("After select, sel_retval: %d\n", sel_retval);
		DEBUG("waittime: %d.%d\n", (int)waittime.tv_sec, (int)waittime.tv_usec);

		// Search for ready senders
		if( sel_retval > 0 ) {
			for(i = 0; i < MAX_CLIENTS; ++i) { // From whom can you read?
				if(FD_ISSET(clients[i].socket, &fd_readready)) {
					// buf needs cleaning -- it's illogical to me
					memset(buf, 0, sizeof(buf));
					read_retval = recv(clients[i].socket, buf, MAX_MSG_LEN, 0);
					DEBUG("%d %s :\n%s\n", clients[i].socket, clients[i].nick, buf);
					if(read_retval > 0) {
						// Sanitize
						//if( buf[read_retval-1] == '\n' ) buf[read_retval-1] = '\0';
						//clear_buf(buf);
						strncpy(cmdbuf, buf, MAX_CMD_LEN);
						cmdbuf[MAX_CMD_LEN] = '\0';

						if(strcmp(cmdbuf, "/nick") == 0) { set_nick(i, buf); continue; }
						if(strcmp(cmdbuf, "/clos") == 0) { close_client(i); continue; }
						if(strcmp(cmdbuf, "/file") == 0) { recv_file(i, buf, read_retval); continue; }

						if(clients[i].nick[0] != '\0')
							send_to_clients(i, buf, read_retval);
						else
							tell_to_give_a_nick(i);

					} else {
						close_client(i);
					}
				}
			}
		}
	}
}

void* listeningthread(void* param) {
	struct sockaddr_in listenaddr, remoteaddr;
	unsigned int remoteaddr_len;
	int listensock;
	int i, truth = 1;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	// The PF = Internet Family, 0 - autodefault for INET/STREAM = TCP 
	if((listensock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		ERROR("Cannot create a socket (%d).\n", errno);
		return NULL;
	}

	// Rapid local address & port reuse
	if((setsockopt(listensock, SOL_SOCKET, SO_REUSEADDR, &truth, sizeof(int))) == -1) {
		ERROR("Cannot set socket options (%d).\n", errno);
		close(listensock);
		exit(1);
	}

	// For my own safety, clean the structure
	memset(&listenaddr, 0, sizeof(listenaddr));
	// Init the localaddr structure
	listenaddr.sin_family = AF_INET;
	// There is only the s_addr member in the sin_addr structure
	listenaddr.sin_addr.s_addr = INADDR_ANY;
	// Local to network byte order
	//listenaddr.sin_port = htons(LISTEN_PORT);
	listenaddr.sin_port = htons(cmdport);

	// Bind a name to a socket (ID: IP, port)
	if(bind(listensock, (struct sockaddr*)&listenaddr, sizeof(listenaddr)) == -1) {
		ERROR("Cannot bind (%d).\n", errno);
		close(listensock);
		exit(1);
	}

	// Tell the sock it wants to accept incoming conns.
	if((listen(listensock, MAX_QUEUED_CONNS)) == -1) {
		ERROR("Cannot start listening (%d).\n", errno);
		close(listensock);
		exit(1);
	}

	while(1) {
		// Find an empty slot for a client - cleaning may occur
		SEM_P(sec_clients);
			for(i = 0; i < MAX_CLIENTS && clients[i].socket != 0; ++i);
		SEM_V(sec_clients);
		// An empty slot found
		if(i < MAX_CLIENTS) {
			DEBUG("client index: %d\n", i);
			// Accept the incoming subspace signal
			// accept substitutes the structure's size
			remoteaddr_len = sizeof(remoteaddr);
			clients[i].socket =
				accept(listensock, (struct sockaddr*)&remoteaddr, &remoteaddr_len);
			clients[i].addr = remoteaddr.sin_addr;
			clients[i].port = remoteaddr.sin_port;
			setmaxsock(clients[i].socket);
			DEBUG2("client %d for socket: %d\n", i, clients[i].socket);
		} else {
			DEBUG("no free slots for clients\n");
		}
	}

	close(listensock);
}

void siginthandler(int s) {
	DEBUG2("SIGINT: terminating threads...\n");
	pthread_cancel(datathread_id);
	pthread_cancel(listeningthread_id);
}

void* timethread(void* param) {
	int i;
	while(1) {
		sleep(10);
		for(i = 0; i < MAX_CLIENTS; ++i)
			clients[i].written >>= 1;
	}
}

int main(int argc, char **argv) {
	int one = 1;

	if(argc != 2 && argc != 3) {
		printf("%s PORT\n", argv[0]);
		return 1;
	}

	if(argc == 2)
		logfile = fopen("log.txt", "a");
	else
		logfile = fopen(argv[2], "a");

	if(!logfile) {
		ERROR("Cannot open logfile.\n");
		exit(1);
	}

	sscanf(argv[1], "%d", &cmdport);

	if((sec_clients = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600)) == -1) {
		ERROR("Cannot get a semaphore.\n");
		fclose(logfile);
		exit(1);
	}

	if((semctl(sec_clients, 0, SETVAL,
#ifdef SUNCOMPILATION
&
#endif
			one)) == -1) {
		ERROR("Cannot set a semaphore.\n");
		semctl(sec_clients, 0, IPC_RMID, NULL);
		fclose(logfile);
		exit(1);
	}

	// Initial cleanup of client structures
	memset(clients, 0, sizeof(struct client)*MAX_CLIENTS);

	// End the program with ^C
	signal(SIGINT, siginthandler);

	pthread_create(&listeningthread_id, NULL, listeningthread, NULL);
	pthread_create(&datathread_id, NULL, datathread, NULL);
	pthread_create(&timethread_id, NULL, timethread, NULL);

	pthread_join(listeningthread_id, NULL);
	//pthread_join(datathread_id, NULL);

	semctl(sec_clients, 0, IPC_RMID, NULL);
	fclose(logfile);

	return 0;
}
